#!/usr/bin/perl -w
#
# cfgtocdb - parses configuration files and translates them into cdbmake
# record format 
#
# see CFG_FORMAT_CDB for information on formatting
#

use strict;

use vars qw(%modname %port %modset %data %index $count %modsetref %modnameref);
# index: keys that have thus far been used
# count: which arbitrary number we're up to
# modsetref:  name->key mapping for modsets
# modnameref: name->key mapping for modnames

$count = 0;

while (<>) {
    /^\..*:/       && do { parse_modname($_); next; };
    /^[+-][0-9]*:/ && do { parse_port   ($_); next; };
    /^=.*:/        && do { parse_modset ($_); next; };
    /^~.*:/        && do { parse_data   ($_); next; };
}

foreach my $k (keys(%modname)) {
#    print ".$k: " , $modname{$k}->[0] , "\n";
    foreach my $v (@{$modname{$k}}) {
        print "+5," . length($v) . ":." . pack("L", $k) . "->$v\n";
    }
}

foreach my $k (keys(%modset)) {
#    print ".$k: " , $modset{$k}->[0] , "\n";
#    foreach my $ent (@{$modset{$k}}) {
#        print "=$k: " , $ent , "\n";
#    }
    foreach my $v (@{$modset{$k}}) {
        print "+5," . length($v) . ":=" . pack("L", $k) . "->$v\n";
    }
}

foreach my $k (keys(%port)) {
#    print substr($port{$k}->[0],0,1) . "$k: " , $port{$k}->[0] , "\n";
    foreach my $v (@{$port{$k}}) {
        print "+4," . length($v) . ":" . pack("L", $k) . "->$v\n";
    }
}

foreach my $k (keys(%data)) {
    foreach my $v (@{$data{$k}}) {
        print "+" . length($k) . "," . length($v) . ":$k->$v\n";
    }
}

print "\n";

###############
# sanity checks
###############

# ckref \%refset, \%defset, $reftype
#
# checks all reference entries in %refset, making sure
# that they're defined in %modset.  $reftype is an informational
# description (string) for use in the error message

sub ckref {
    my ($refset, $defset, $reftype) = @_;
    foreach my $ref (keys %{$refset}) {
        my ($err) = 0;
        if (!defined $defset->{$refset->{$ref}}) {
            print STDERR "module name reference '$ref' is not defined.\n";
            $err = 1;
        }
        exit 1 if $err;
    }
}

ckref(\%modnameref, \%modname, "module name");
ckref(\%modsetref,  \%modset,  "module set");


# all ports are numerical
#
# $port{key} = [
#       [
#           act:8,
#           dbits:8*4, dip:8*4, dport:32, dportmask:32,
#           sbits:8*4, sip:8*4, sport:32, sportmask:32,
#           modulekey:32
#       ], ...
#   ]
#
# $modset{key} = [
#       [
#           howmany:8,
#           modulename:32,
#           ...<howmany times>
#           datalen:16,
#           data:<datalen>
#       ], ...
#   ]
#
# $modname{key} = [
#       [
#           pathlen:16,
#           path:<pathlen>
#       ], ...
#   ]
#
# $dataname{key} = [
#       [
#           # datalen is cdb value length
#           data
#       ], ...
#   ]

# getref $key, \%refset
#
# looks up $key in %refset and returns the value.
# if the entry doesn't exist, it will first be created with a unique value.
sub getref {
    my ($key, $refset) = @_;
    $refset->{$key} = ++$count
        unless defined $refset->{$key};

    return $refset->{$key};
}

# malformed $info
#
# exits with an error message of $info; reports current input line number

sub malformed {
    print STDERR "$.: @_\n";
    exit 1;
}

# chimp [$_]
#
# removes the first character in $_
# returns the character removed if successful
# returns undef if no character was found to be removed

sub chimp {
    s/^(.)// && return $1;
    return undef;
}

sub parse_modname {
    chimp; chomp;

    my ($namekey);
    my ($name, $path) = split (/:/, $_, 2);
    format_error() unless
        (defined($name) && defined($path));

    $namekey = getref($name, \%modnameref);

#    print "ref $name is $modnameref{$name}\n";
    #$modname{$namekey} = [ $path ];
    push(@{$modname{$namekey}}, $path);
}

sub parse_modset {
    chimp;
    my ($setname, @fields) = split (/:/);
    my (@modnames);
    while (@fields) {
        my ($modname) = shift @fields;
        last if $modname eq "";

        my ($namekey) = getref($modname, \%modnameref);
#        print "ref $modname is $modnameref{$modname}\n";
        push (@modnames, $namekey);
    }

    malformed ("too many module entries") if (@modnames > 255);

    my ($data, $datalen);
    if ($fields[0] =~ /^[0-9]{1,4},/) {
        ($datalen, $data) = split(/,/, $fields[0], 2);
        shift @fields;
    }
    $data .= join (':', @fields) if @fields;

    if (defined($datalen) && length($data) < $datalen) {
        read (STDIN, substr($data, length($data), 0), $datalen - length($data))
            or malformed("data needs at least $datalen bytes");
        my ($tmp) = <>;
        chomp($tmp);
        $tmp eq "" || malformed("garbage after data");
    } elsif (defined($datalen)) {
        substr($data,$datalen) = "";
    } else {
        chomp ($data);
    }

    my ($setkey);
    if ($setname eq "") {
      $setkey = 0;
    } else {
      $setkey = getref($setname, \%modsetref);
    }

    my ($value) = pack("CL" . @modnames . "Sa*",
            scalar(@modnames), @modnames, length($data), $data);
    push(@{$modset{$setkey}}, $value);
}

sub parse_port {
    my ($act) = chimp; chomp;
    my ($dpt, $spt, $sip, $dip, $setname) = split (/:/, $_, 5);
    my ($sipmask,    $dipmask, $sptmask, $dptmask);
    my (@sipmaskset, @dipmaskset, @sipset, @dipset);

    malformed("bad port") if ($dpt ne "" && $dpt !~ /^[0-9]+$/);

    ($spt, $sptmask) = split('/', $spt, 2);
    ($dpt, $dptmask) = split('/', $dpt, 2);
    ($sip, $sipmask) = split('/', $sip, 2);
    ($dip, $dipmask) = split('/', $dip, 2);

    $dpt = 0, $dptmask = 0 if (!defined($dpt) || $dpt eq "");
    $spt = 0, $sptmask = 0 if (!defined($spt) || $spt eq "");
    $dptmask = 0xffffffff  if (!defined($dptmask));
    $sptmask = 0x00000000  if (!defined($sptmask));

    $sip = "0.0.0.0", $sipmask = "0.0.0.0" if (!defined($sip) || $sip eq "");
    $dip = "0.0.0.0", $dipmask = "0.0.0.0" if (!defined($dip) || $dip eq "");

    $sipmask = "255.255.255.255" if (!defined($sipmask));
    $dipmask = "255.255.255.255" if (!defined($dipmask));

    @sipset     = split(/\./, $sip);
    @dipset     = split(/\./, $dip);
    @sipmaskset = split(/\./, $sipmask);
    @dipmaskset = split(/\./, $dipmask);

    malformed("bad ip ($sip)") if (@sipset != 4);
    malformed("bad ip ($dip)") if (@dipset != 4);
    malformed("bad mask ($sipmask)") if (@sipmaskset != 4);
    malformed("bad mask ($dipmask)") if (@dipmaskset != 4);

    foreach my $val (@sipset, @dipset) {
        malformed("bad ip octet ($val)")
            if (!($val =~ /^[0-9]+$/ && $val ge 0 && $val le 255));
    }

    my ($setkey) = 0;
    $setkey = getref($setname, \%modsetref)
        if (defined $setname && $setname ne "");


    my ($value) = pack ("C4C8LLC8LLL",
        ord($act), 0, 0, 0, @dipmaskset, @dipset, $dpt, $dptmask,
              @sipmaskset, @sipset, $spt, $sptmask, $setkey);

    push(@{$port{$dpt}}, $value);
}

sub parse_data {
    chomp;

    my ($dataname, $data) = split (/:/, $_, 2);
    malformed("bad data format")
        if (!defined($dataname) || !defined($data));

    push(@{$data{$dataname}}, $data);
}
