#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <cdb.h>
#include "lxmisc.h"
#include "misc.h"
#include "data.h"
#include "naw_log.h"


/***************************************************************************
  match_entry

  matches info to gathered entry slot, taking masks and such into account.

  returns 0 on successful match
  returns -1 otherwise
 ***************************************************************************/

static inline int match_entry(info, entry)
    struct netinfo *info;
    struct netinfo *entry;
{
    NAW_LOG (NAW_DEBUG3, LXM_q("comparing address data ("),
            LXM_x(info->dst.host),
            LXM_q("/"),
            LXM_x(entry->dst.host),
            LXM_q(" : "),
            LXM_x(entry->dst.bits),
            LXM_q(")"));

    /* dest port is guaranteed, since it's the key.
     * XXX: If it stops being guaranteed, add a check here.
     * XXX: remember that 0 is intended to be the catch-all.
     */
    if (cmp_host_bits (info->dst.host, entry->dst.host, entry->dst.bits, 0))
        return -1;

    if (cmp_host_bits (info->src.host, entry->src.host, entry->src.bits, 0))
        return -1;

    /* source port may use masks */
    if (cmp_host_bits (info->src.port, entry->src.port, entry->src.pits, 0))
        return -1;

    return 0;
}

/***************************************************************************
  judge_entry

  performs granular checks on configuration entries
  [entry] is a value (possibly one of many) for the key info->dst->port.

  returns [sd] if the connection should be accepted
  returns  -1  if the connection should be refused.

  XXX: differentiate between errors and rejections
 ***************************************************************************/

static inline int judge_entry (sd, c, type, info, entry)
    int sd;
    struct cdb *c;
    ub1_t type;
    struct netinfo *info;
    struct netentry *entry;
{
    int y, n, r, v;
    struct cdb_find cf;
    struct naw_data data;
    data.bs = 0;

    switch (entry->action) {
        case '+':
            y = sd;
            n = -1;
            break;
        case '-':
            y = -1;
            n = sd;
            break;
        default:
            NAW_LOG(NAW_NOTICE, LXM_q("config: unknown action requested"));
            return -1;
    }
    r = y;

    if (entry->module_key) {
        char key[5];
        ub1_t once = 0;
        struct naw_handle handle;

        key_load(key, '=', entry->module_key);

        handle.sd    = sd;
        handle.type  = type;
        handle.info  = info;
        handle.entry = entry;

        if (cdb_findinit(&cf, c, key, 5) < 0) {
            NAW_LOG(NAW_NOTICE,LXM_q("config: error initializing cdb"));
            r = -1; goto OUT;
        }

        for (;;) {
            v = cdb_findnext(&cf);
            switch (v) {
                case 1: break;
                case 0: if (once) goto OUT;
                        NAW_LOG(NAW_NOTICE,
                                LXM_q("config: associated module set"
                                    " does not exist"));
                        r = NAW_ERR; goto OUT;
                case -1:NAW_LOG(NAW_NOTICE,LXM_q("config: cdb short read"));
                        r = NAW_ERR; goto OUT;
            }

            once = 1;
            NAW_LOG(NAW_DEBUG3,LXM_q("found associated module set"));

            if (naw_data_init(&data)) { r = NAW_ERR; goto OUT; }
            v = modentry_run(c, &handle, &data);

            switch (v) {
                case NAW_NO:  r = n; goto OUT;
                case NAW_YES: break;
                case NAW_ERR:
                default:
                    r = -1; goto OUT;
            }

            naw_data_free(&data);
            data.bs = 0;
        }
    }

OUT:
    if (data.bs) naw_data_free(&data);
    return r;
}


/***************************************************************************
  naw_verify

  gathers information on the socket at hand, then finds a configuration
  entry for it.  If no matching entry is found, the connection is refused.

  refused connections have errno set to ECONNABORTED.

  returns -1 if the connection is denied or if there is an error
  returns sd if the connection is accepted.

  sd will be closed on error.
  XXX: ensure that errno is always set properly
 ***************************************************************************/

int naw_verify_internal (sd, type, addr)
    int sd;
    ub1_t type;
    void *addr;
{
    struct sockaddr_in *sin = addr;
    struct netinfo sdinfo;
    struct netentry sdentry;
    struct cdb_find cf;
    struct cdb c;
    int cdbfd, r;
    unsigned pos, len;
    ub4_t key;

    {
        struct sockaddr_in tmp_addr;

        if (!sin) {
            socklen_t tmp_addr_len = sizeof (struct sockaddr_in);

            NAW_LOG (NAW_DEBUG, LXM_q("sockaddr passed as null"));

            sin = &tmp_addr;
            if (getpeername (sd, (struct sockaddr *)sin, &tmp_addr_len))
            {
                close (sd);
                return -1;
            }
        }

        /* don't interfere with non-inet sockets */
        if (sin->sin_family != AF_INET) {
            NAW_LOG(NAW_DEBUG2,LXM_q("non-inet socket; ignoring"));
            return sd;
        }

        /* populate the idat structure */
        if (get_addr_data (&sdinfo, sin, type, sd)) {
            close (sd);
            return -1; /* errno is preset */
        }
    }

#define LXM_s(s) LXM_mb(s, strlen(s))

    NAW_LOG(NAW_DEBUG3,
            LXM_q("reading config file: "),
            LXM_s(cdb_config[(int)type]));
    cdbfd = open (cdb_config[(int)type], O_RDONLY);
    if (cdbfd == -1) {
        naw_logfmt(NAW_NOTICE, "cdb config file (%s): %s",
                cdb_config[(int)type], strerror(errno));
        close (sd);
        return -1;
    }


#define ERROUT do {       \
    errno = ECONNABORTED; \
    close(sd);            \
    r = -1; goto OUT;     \
} while (0)


    CDB_INIT(&c, cdbfd);

    key = sdinfo.dst.port;

    if (cdb_findinit(&cf, &c, &key, 4) < 0) {
        NAW_LOG(NAW_NOTICE,LXM_q("error initializing cdb"));
        ERROUT;
    }

    for (;;) {
        /*
           fprintf(stderr, "searching port: %x\n", (unsigned int)key);
           fflush(stderr);
           */
        r = cdb_findnext(&cf);

        if (r == -1) {
            naw_logfmt(NAW_NOTICE, "cdb read failed: %s",strerror(errno));
            ERROUT;
        } else if (!r) {
            if (key) {
                key = 0;  /* port 0 is the catch-all */
                if (cdb_findinit(&cf, &c, &key, 4) < 0) { ERROUT; }
                continue;
            }  

            NAW_LOG(NAW_DEBUG2,
                    LXM_q("no matching configuration entry found"));
            ERROUT;
        }

        pos = cdb_datapos(&c);
        len = cdb_datalen(&c);

        if (len != sizeof(sdentry)) {
                /*
            printf ("len: %d; size: %d\n", (int)len, (int)sizeof(sdentry));
            */
            naw_logfmt(NAW_NOTICE,
                    "config: cdb value is the incorrect size for casting"
                    " (%d bytes)", len);
            continue; /* ignore this entry */
        }

        if (cdb_read(&c, (char *)&sdentry, sizeof(sdentry), pos) == -1) {
            NAW_LOG(NAW_NOTICE,LXM_q("config: cdb short read"));
            ERROUT;
        }

        r = match_entry(&sdinfo, &sdentry.info);
        if (r == -1) continue;

        NAW_LOG(NAW_DEBUG3,LXM_q("found matching configuration entry"));

        r = judge_entry (sd, &c, type, &sdinfo, &sdentry);
        if (r == -1) { close(sd); errno = ECONNABORTED; }
        break;
    }

OUT:
    cdb_free(&c);
    close(cdbfd);
    return r;
}

int naw_verify (sd, type, addr)
    int sd;
    ub1_t type;
    void *addr;
{
    EXPLICIT_INIT(if (!naw_initialized) naw_init());
    return naw_verify_internal(sd, type, addr);
}
