#include <sys/types.h>
#include <netinet/in.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <lx_string.h>
#include <minimisc.h>
#include <cdb.h>

#include "misc.h"

enum stages {
    STAGE_COMMENT=1,
    STAGE_IP,
    THREE,
    END
};

char *stageset[END];

int cfd;
static struct cdb_make db;
blobset *bs = 0;
lx_s key  = {0};
lx_s data = {0};

char *dest, *tmp;
char eofok = 1;
int byte = 0, line = 1;

#define ERROUT(msg) fprintf(stderr,msg" (line %d, byte %d)\n",line,byte)


void
die_cleanup(void)
{
    close(cfd);
    unlink(tmp);
    exit(1);
}

void
die_format(void)
{
    ERROUT("bad config file format");
    die_cleanup();
}

void
die_nomem(void)
{
    ERROUT("out of memory");
    die_cleanup();
}

void
die_write(void)
{
    perror("write");
    die_cleanup();
}

void
die_finish(void)
{

    if (cdb_make_finish(&db))
        die_write();

    close(cfd);
    if (rename(tmp, dest)) {
        perror(dest);
        die_cleanup();
    }
    exit(0);
}

char
nextc(void)
{
    int c;
    c = fgetc(stdin);
    if (c == EOF) {
        if (!feof(stdin)) {
            ERROUT("read error");
            die_cleanup();
        }
        if (eofok)
            die_finish();

        ERROUT("unexpected eof");
        die_cleanup();
    }

    byte++;
    if (c == '\n') line++;
    return (char) c;
}

ub4_t getref(s, len)
    char *s;
    unsigned len;
{
    void *bv;
    static ub4_t count = 0;

    if (!bs) bs = blob_new(16, 0);
    if (!bs) die_nomem();

    bv = blob_get(bs, s, len);
    if (bv) return *(ub4_t *)bv;

    bv = calloc(1, sizeof(ub4_t));
    *(ub4_t *)bv = ++count;
    if (!bv) die_nomem();
    if (blob_register(bs, s, len, bv, 0))
        die_nomem();

    return *(ub4_t *)bv;
}

void
addkey(void)
{
    if (cdb_make_add(&db, key.s, key.len, data.s, data.len))
        die_write();
}

#define IS_DIGIT(c) (c >= '0' && c <= '9')

void
ignore_line(void)
{
    do {} while (nextc() != '\n');
}

void
slurp_line(lx_s *s)
{
    char c;
    for (;;) {
        c = nextc();
        if (c == '\n') return;
        lx_cadd(s, c);
    }
}

void
slurp_seg(lx_s *s)
{
    char c;
    for (;;) {
        c = nextc();
        if (c == ':') return;
        if (c == '\n') die_format();
        lx_cadd(s, c);
    }
}

unsigned
add_digit(unsigned n, char digit)
{
    n *= 10;
    n += digit - '0';
    return n;
}

void
slurp_quad(ub4_t *dest, char *next)
{
    char c;
    ub1_t octets=4;
    ub1_t thisoctet=1;
    unsigned ov = 0;
    
    *dest = 0;

    for (;;) {
        c = nextc();

        if (c == '.') {
            *dest |= ov << ((octets - thisoctet) * 8);
            ov = 0;

            if (++thisoctet > octets) die_format();
            continue;
        }

        if (c == '/' || c == ':')
            break;

        if (!IS_DIGIT(c)) die_format();
        ov = add_digit(ov, c);
        if (ov > 0xff) die_format();
    }

    *dest = htonl(*dest | ov << ((octets - thisoctet) * 8));
    *next = c;
}

void
slurp_ip(ub4_t *ip, ub4_t *netmask)
{
    char c;
    *ip = 0;
    *netmask = 0xffffffff;

    slurp_quad(ip, &c);
    if (c == ':') {
        if (!*ip) *netmask = 0;
        return;
    }

    if (c != '/')
        die_format();

    slurp_quad(netmask, &c);
    if (c != ':')
        die_format();
}

void
slurp_port(ub4_t *port, ub4_t *netmask)
{
    char c;

    /* defaults:
     * 0 port, 0 netmask
     * if port becomes nonzero, default is full netmask.
     */
    *port = 0;
    *netmask = 0;

    for (;;) {
        c = nextc();
        if (c == ':') break;
        if (c == '/') {
            /* 0 port gets 0 default netmask */
            slurp_quad(netmask, &c);
            if (c == ':') break;
            die_format();
        }

        if (!IS_DIGIT(c))
            die_format();
        *port = add_digit(*port, c);
        if (*port) *netmask = 0xffffffff;
    }
    *port = htons(*port);
}

void
parse_port(char c)
{
    lx_s setname = {0};
    struct netentry entry;

    eofok = 0;

    memset(&entry, 0, sizeof(entry));
    entry.action = c;

    slurp_port(&entry.info.dst.port,
               &entry.info.dst.pits);
    slurp_port(&entry.info.src.port,
               &entry.info.src.pits);

    slurp_ip(&entry.info.dst.host,
             &entry.info.dst.bits);
    slurp_ip(&entry.info.src.host,
             &entry.info.src.bits);

    slurp_line(&setname);
    if (setname.s)
        entry.module_key = getref(setname.s, setname.len);

    eofok = 1;

    if (lx_striset(&key, (char *)&entry.info.dst.port, 4))
        die_nomem();
    if (lx_striset(&data,(char *)&entry, sizeof(entry)))
        die_nomem();

    addkey();

    lx_free(&setname);
}

void
parse_modset(char c)
{
    lx_s name = {0};
    lx_s modset = {0};
    lx_s modata = {0};
    lx_s digits = {0};
    ub4_t nameref, ref = 0;
    ub1_t count;
    ub2_t datalen;

    unsigned datsize = 0;

    eofok = 0;

    slurp_seg(&name);
    if (!name.s) nameref = 0;
    else         nameref = getref(name.s, name.len);

    for (count=0;;) {
        name.len = 0;
        slurp_seg(&name);
        if (!name.len) break;

        if (count == (ub1_t)-1)
            die_format();

        count++;
        ref = getref(name.s, name.len);
        if (lx_striadd(&modset, (char *)&ref, 4)) die_nomem();
    }

    digits.len = 0;
    for (;;) {
        c = nextc();
        if (c == ',' && digits.len > 0)
            break;

        /* hard max: 999999 */
        if (!IS_DIGIT(c) || digits.len >= 6) {
            if (lx_striset(&modata, digits.s, digits.len)) die_nomem();
            if (c != '\n') {
                lx_cadd(&modata, c);
                slurp_line(&modata);
            }
            datsize = 0;
            break;
        }

        datsize = add_digit(datsize, c);
        if (lx_cadd(&digits, c)) die_nomem();
    }

    lx_free(&digits);

    while (datsize--)
        if (lx_cadd(&modata, nextc())) die_nomem();

    eofok = 1;

    if (modata.len > (ub2_t)-1)
        die_format();
    datalen = (ub2_t) modata.len;

    if (lx_strset (&key, "=")) die_nomem();
    if (lx_striadd(&key, (char *)&nameref, sizeof(ub4_t))) die_nomem();
    if (lx_striset(&data, (char *)&count, 1)) die_nomem();
    if (lx_striadd(&data, modset.s, modset.len)) die_nomem();
    if (lx_striadd(&data, (char *)&datalen, sizeof(ub2_t))) die_nomem();
    if (lx_striadd(&data, modata.s, datalen)) die_nomem();

    addkey();

    lx_free(&name);
    lx_free(&modset);
}

void
parse_data(char c)
{
    lx_s name  = {0};
    lx_s value = {0};
    /*ub4_t ref;*/

    eofok = 0;

    slurp_seg(&name);
    if (!name.s) die_format();
    /*
    ref = getref(name.s, name.len);
    */
    
    slurp_line(&value);
    if (!value.s) die_format();

    eofok = 1;

    if (lx_strset (&key, "~")) die_nomem();
    if (lx_striadd(&key, name.s, name.len)) die_nomem();
    /*
    if (lx_striadd(&key, (char *)&ref, sizeof(ub4_t))) die_nomem();
    */
    if (lx_striset(&data, value.s, value.len)) die_nomem();

    addkey();

    lx_free(&name);
    lx_free(&value);
}

void
parse_modname(char c)
{
    lx_s name = {0};
    lx_s path = {0};
    ub4_t ref;

    eofok=0;

    slurp_seg(&name);
    ref = getref(name.s, name.len);
    slurp_line(&path);

    eofok=1;

    if (lx_strset (&key, ".")) die_nomem();
    if (lx_striadd(&key, (char *)&ref, sizeof(ub4_t))) die_nomem();
    if (lx_striset(&data, path.s, path.len)) die_nomem();

    addkey();

    lx_free(&name);
    lx_free(&path);
}

#define WHITESPACE " \t\r\n"

int
main(argc, argv)
    int argc;
    char **argv;
{
    char c;

    if (argc != 3) {
        fprintf(stderr, "usage: naw-data dest tmp <input\n");
        exit (1);
    }

    dest = argv[1];
    tmp  = argv[2];

    cfd = open(tmp, O_WRONLY|O_NDELAY|O_TRUNC|O_CREAT, 0644);
    if (cfd == -1) {
        perror (tmp);
        exit(1);
    }

    if (cdb_make_start(&db, cfd))
        die_write();

    for (;;) {
        c = nextc();

        if (memchr(WHITESPACE, c, sizeof(WHITESPACE)-1))
            continue;
        if (c == '+' || c == '-')
            parse_port(c);
        else if (c == '=')
            parse_modset(c);
        else if (c == '~')
            parse_data(c);
        else if (c == '.')
            parse_modname(c);
        else 
            ignore_line();
    }

    return 0;
}

