#ifndef LIBNAW_GENLINK_H
#define LIBNAW_GENLINK_H


/***************************************************************************
  This file contains general macros for creating linked list structures.

  Each link is one-dimensional.
 ***************************************************************************/


/***************************************************************************
  analogous prototypes:
    void GENLINK_new (type *link, [type]);
    void GENLINK_add (type **firstlinkp, const type *newlink, [next], [type]);
    void GENLINK_remove (type **firstlinkp, const type *rmlink, [next], [type]);
    void GENLINK_destroy (type *link);

    future, if necessary:
      void GENLINK_find (const type *firstlink, type **linkp, [matchexp], [next], [type]);

  sample use:
    extern struct link *firstlink;
    ...
    struct link *link;

    GENLINK_new(link,struct link);
    GENLINK_add(&firstlink,link,next,struct link);
    ...
    {
        struct link *rmlink = firstlink;
        while (rmlink) {
            if ([matchexp]) break;
            rmlink = rmlink->next;
        }
        if (rmlink) {
            GENLINK_remove(&firstlink,link,next,struct link);
            GENLINK_destroy(link);
        }
    }

  Notes: next is the element name within type that points to the next link.
 ***************************************************************************/

#define GENLINK_new(link,type) do { \
    link = malloc (sizeof(type)); \
    if (!(link)) die_nomem(); \
    bzero (link, sizeof(type)); \
} while (0)


#define GENLINK_add(firstlinkp,newlink,next,type) do { \
    type *l = 0, *i = *firstlinkp; \
    while (i) { \
	l = i; \
	i = i->next; \
    } \
    if (!l) *firstlinkp = newlink; \
    else        l->next = newlink; \
} while (0)

#define GENLINK_remove(firstlinkp,badlinkp,matchexp,next,type) do { \
    type *l = 0, *badlink; \
    badlink = *firstlinkp; \
    while (badlink && !(matchexp)) { \
	l = badlink; \
	badlink = badlink->next; \
    } \
    if (!badlink) { *badlinkp = badlink; break; } \
    l->next = badlink->next; \
} while (0)

#define GENLINK_destroy(link) do { \
    free (link); \
} while (0)

#endif
